@@ -2,6 +2,7 @@ import logging
 import os
 import tempfile
 import shutil
+
 from . import pyjson
 
 
@@ -20,30 +21,44 @@ class BaseStateCommitter(object):
 
 class JsonStateCommitter(BaseStateCommitter):
     def __init__(self, json_path):
-        self.temp_file = None
         self._state_file = json_path
 
+    @property
+    def path(self):
+        return self._state_file
+
+    @property
+    def state_file_exists(self) -> bool:
+        return os.path.isfile(self._state_file)
+
     def load(self, default=None):
+        """
+        Loads and reads from state file. If an error occurs, `default` will be returned, if it exists.
+        """
         try:
             with open(self._state_file) as json_data:
                 return pyjson.load(json_data)
-        except (ValueError, EnvironmentError):
-            return dict() if default is None else default
+        except (ValueError, EnvironmentError) as e:
+            if default is not None:
+                logger.warning(
+                    f"Failed to load state from {self.path} due to {e}. Returning default: {default}.",
+                    exc_info=True,
+                )
+                return default
+            logger.error(f"Failed to load state from {self.path}: {e}", exc_info=True)
+            raise e
 
     def commit(self, state):
         """ Write out the state in JSON format to a temporary file and rename it into place """
         with tempfile.NamedTemporaryFile(delete=False) as f:
             try:
-                self.temp_file = f.name
                 f.write(pyjson.dumps(state).encode())
                 f.flush()
-                shutil.copyfile(self.temp_file, self._state_file)
+                shutil.copyfile(f.name, self._state_file)
             finally:
                 try:
-                    os.unlink(self.temp_file)
+                    os.unlink(f.name)
                 except FileNotFoundError:
                     logger.error(
-                        "Problem occurred in deleting temp file {} via os.unlink".format(
-                            self.temp_file
-                        )
+                        "Problem occurred in deleting temp file {} via os.unlink".format(f.name)
                     )
